/*
 * ============================================================================
 * History
 * =======
 * 29 Dec 2015 : Created
 *
 * Copyright (C) Bridgetek Pte Ltd
 * ============================================================================
 *
 * This source code ("the Software") is provided by Bridgetek Pte Ltd 
 * ("Bridgetek") subject to the licence terms set out
 * http://brtchip.com/BRTSourceCodeLicenseAgreement/ ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of Bridgetek liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on Bridgetek's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, Bridgetek
 * has no liability in relation to those amendments.
 * ============================================================================
 */

#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include "ft900.h"
#include "tinyprintf.h"

/* lwIP Headers. */
#include "lwip/sockets.h"
#include "lwip/ip4_addr.h"
#include "lwip/apps/mqtt.h"

/* netif Abstraction Header. */
#include "net.h"
#include "netdb.h"

/* FreeRTOS Headers. */
#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include "semphr.h"

#define DEBUG
#ifdef DEBUG
#define DEBUG_PRINTF(...) do {CRITICAL_SECTION_BEGIN;tfp_printf(__VA_ARGS__);CRITICAL_SECTION_END;} while (0)
#else
#define DEBUG_PRINTF(...)
#endif

#ifdef MBEDTLS_DEBUG_C
#define MBEDTLS_DEBUG_LEVEL 4
#endif

#define STR_HELPER(x) #x
#define STR(x) STR_HELPER(x)

/* Two Demo Tasks are implemented here to illustrate the usage of lwIP using static/dynamic IP address.
 *
 * SERVER DEMO
 * ------------------------------
 * A TCP/UDP server runs listening for requests on PORT 80. Received data is processed
 * and then a response message is sent. The socket is then closed.
 *
 * CLIENT DEMO
 * ------------------------------
 * A TCP/UDP client tries to connect to a predefined server IP address. Once connection is successful, a message
 * is sent and the response from the server is received. The socket is then closed.
 *
 * Both demos work with the Python test scripts to be run on host PC.
 *
 */

/* Default network configuration. */
#define NET_USE_DHCP            1     // 1: Dynamic IP, 0: Static IP
#define NET_DEFAULT_IP_ADDR     IP_ADDR_ANY
#define NET_DEFAULT_GW_ADDR     IP_ADDR_ANY
#define NET_DEFAULT_MASK        IP_ADDR_ANY

/*
 * NOTE: Be sure to update MQTT_BROKER to the correct Host PC IP address.
 */
#define MQTT_CLIENT_NAME              "lwip_test"
#define MQTT_BROKER                   "192.168.1.100"
#define MQTT_BROKER_PORT              MQTT_PORT

/* FreeRTOS task handles. */
static TaskHandle_t gx_TaskConnect;
static TaskHandle_t gx_TaskMQTT;
static TaskHandle_t gx_TaskMonitor;

/* Task configurations. */
#define TASK_MQTT_STACK_SIZE          (2000)          //Task Stack Size
#define TASK_MQTT_PRIORITY            (1)             //Task Priority
#define TASK_CONNECT_STACK_SIZE       (200)           //Task Stack Size
#define TASK_CONNECT_PRIORITY         (1)             //Task Priority
#define TASK_MONITOR_STACK_SIZE       (200)           //Task Stack Size
#define TASK_MONITOR_PRIORITY         (1)             //Task Priority

/* Polling settings. */
#define MQTT_CONNECT_POLL_MS          (100)
#define MQTT_CONNECT_RETRY_MS         (2500)
#define TLS_RECONNECT_DELAY_MS        (1000)
#define MONITOR_PERIOD_MS             (1000)

/* Number of pending request on server's listening port. */
#define SERVER_LISTEN_BACKLOG_SIZE      5

/* Debug LwIP resource free in TIME-WAIT state. */
#define DEBUG_LWIP_RESOURCE_FREE        0


#define TOPIC_CONTROL "control"
#define TOPIC_COUNTER "counter"

static void app_ethif_status_cb(int netif_up, int link_up);

void vTaskClient(void *pvParameters);
void client(void);
void vTaskConnect(void *pvParameters);
void vTaskMonitor(void *pvParameters);

struct pub_info {
	const char *topic;
	char message[256];
	size_t len;
};

struct sub_info {
	const char *topic;
	char message[256];
	size_t len;
};

/** tfp_printf putc
 *  @param p Parameters
 *  @param c The character to write */
void myputc(void* p, char c) {
	uart_write((ft900_uart_regs_t*) p, (uint8_t) c);
}

int main(void)
{
	sys_reset_all();
	interrupt_disable_globally();
	/* enable uart */
	sys_enable(sys_device_uart0);
	gpio_function(48, pad_func_3);
	gpio_function(49, pad_func_3);

	uart_open(UART0, 1,
			UART_DIVIDER_115200_BAUD, uart_data_bits_8, uart_parity_none,
			uart_stop_bits_1);
	/* Enable tfp_printf() functionality... */
	init_printf(UART0, myputc);

	/* Set up Ethernet */
	sys_enable(sys_device_ethernet);

	/* Set up I2C */
	sys_enable(sys_device_i2c_master);

	/* Setup I2C channel 0 pins */
	/* Use sys_i2c_swop(0) to activate. */
	gpio_function(44, pad_i2c0_scl); /* I2C0_SCL */
	gpio_function(45, pad_i2c0_sda); /* I2C0_SDA */

	/* Setup I2C channel 1 pins for EEPROM */
	/* Use sys_i2c_swop(1) to activate. */
	gpio_function(46, pad_i2c1_scl); /* I2C1_SCL */
	gpio_function(47, pad_i2c1_sda); /* I2C1_SDA */

	/* Print out a welcome message... */
	uart_puts(UART0,
			"\x1B[2J" /* ANSI/VT100 - Clear the Screen */
			"\x1B[H" /* ANSI/VT100 - Move Cursor to Home */
			"Copyright (C) Bridgetek Pte Ltd \r\n"
			"--------------------------------------------------------------------- \r\n"
			"Welcome to Free RTOS LWIP mbedTLS MQTT Example... \r\n"
			"\r\n"
			"Demonstrate a TCP/UDP Client using the LWIP Stack running on FreeRTOS \r\n"
			"--------------------------------------------------------------------- \r\n");

	if (xTaskCreate(vTaskClient,
			"Client",
			TASK_MQTT_STACK_SIZE,
			NULL,
			TASK_MQTT_PRIORITY,
			&gx_TaskMQTT) != pdTRUE) {
		DEBUG_PRINTF("Client failed\r\n");
	}

	DEBUG_PRINTF("Starting Scheduler.. \r\n");
	/* Start the scheduler so the created tasks start executing. */
	vTaskStartScheduler();

	DEBUG_PRINTF("Should never reach here!\r\n");

	for (;;)
		;
}

/**
 * Task to provide sensor or status readings to the vTaskClient task.
 * This task should be modified to perform any actions required to
 * either send
 */
void vTaskMonitor(void *pvParameters)
{
	int counter = 1;
	struct pub_info pinfo;
	struct sub_info *sinfo;
	uint32_t ulNotifiedValue;

	DEBUG_PRINTF("Task %s started.\r\n", __FUNCTION__);

	pinfo.topic = TOPIC_COUNTER;

	while (1)
	{
		vTaskDelay(pdMS_TO_TICKS(MONITOR_PERIOD_MS));

		if (xTaskNotifyWait(0, 0, &ulNotifiedValue, pdMS_TO_TICKS(0)))
		{
			sinfo = (struct sub_info *)ulNotifiedValue;
			if (sinfo)
			{
				char *end;
				size_t len;

				len = sinfo->len;
				if (len > sizeof(sinfo->message))
				{
					len = sizeof(sinfo->message);
				}
				end = (char *)(sinfo->message) + sinfo->len;
				counter = strtoul((char *)sinfo->message, &end, 0);
			}
		}

		tfp_sprintf(pinfo.message, "%d", counter);
		pinfo.len = strlen(pinfo.message);

		xTaskNotify(gx_TaskMQTT, (uint32_t)&pinfo, eSetValueWithOverwrite);
		counter++;
	}

	DEBUG_PRINTF("Task %s finished.\r\n", __FUNCTION__);
}

/**
 * This callback is enabled when a connection or status change is made
 * to the netif interface within lwIP.
 * Handlers are in place for NETIF initialisation completed and LINK
 * Up/Down.
 * This is purely used for reporting. The vTaskClient task will query
 * the status of the link and netif during it's operation. There is no
 * need to use this callback for program control.
 */
static void app_ethif_status_cb(int netif_up, int link_up)
{
	static int netif_status = 0;
	static int link_status = 0;

	if (netif_up && !netif_status) {
		DEBUG_PRINTF("NETIF UP\r\n");
	}

	if (link_up != link_status)
	{
		if (link_up) {
			DEBUG_PRINTF("LINK UP\r\n");
		}
		else {
			DEBUG_PRINTF("LINK DOWN\r\n");
		}
	}

	netif_status = netif_up;
	link_status = link_up;
}

void vTaskConnect(void *pvParameters)
{
	DEBUG_PRINTF("Task %s started.\r\n", __FUNCTION__);

	while (1)
	{
		// Check for ethernet disconnection.
		if (net_is_link_up())
		{
			if (!ethernet_is_link_up())
			{
				net_set_link_down();
				DEBUG_PRINTF("Ethernet disconnected.\r\n");
			}
		}
		else
		{
			if (ethernet_is_link_up())
			{
				net_set_link_up();
				DEBUG_PRINTF("Ethernet connected.\r\n");
			}
		}
		for (int i = 0; i < 10; i++)
		{
			net_tick();/* 10 ms delay */
			vTaskDelay(1 * portTICK_PERIOD_MS);
		}
	}

	DEBUG_PRINTF("Task %s finished.\r\n", __FUNCTION__);
}

#ifdef MBEDTLS_DEBUG_C
static void mbedtls_debug(void *ctx, int level,
		const char *file, int line,
		const char *str)
{
	(void)level;
	DEBUG_PRINTF("%s:%04d: %s", file, line, str);
}
#endif // MBEDTLS_DEBUG_C

static void client_mqtt_connection_cb(mqtt_client_t *client, void *arg, mqtt_connection_status_t status);
err_t client_mqtt_do_connect(mqtt_client_t *client);
static void client_mqtt_sub_request_cb(void *arg, err_t result);
static void client_mqtt_incoming_publish_cb(void *arg, const char *topic, u32_t tot_len);
static void client_mqtt_incoming_data_cb(void *arg, const u8_t *data, u16_t len, u8_t flags);
static void client_mqtt_pub_request_cb(void *arg, err_t result);
err_t client_mqtt_publish(mqtt_client_t *client, void *arg);
void client_mqtt(void);

static void client_mqtt_connection_cb(mqtt_client_t *client, void *arg, mqtt_connection_status_t status)
{
	if (status == MQTT_CONNECT_ACCEPTED)
	{
		DEBUG_PRINTF("MQTT CONNECT\r\n");
	}
	else
	{
		DEBUG_PRINTF("MQTT CONNECT result: %d\r\n", status);

		vTaskDelay(pdMS_TO_TICKS(MQTT_CONNECT_RETRY_MS));
		/* Try again to connect. */
		client_mqtt_do_connect(client);
	}
}

err_t client_mqtt_do_connect(mqtt_client_t *client)
{
	struct mqtt_connect_client_info_t ci;
	err_t err;
	ip_addr_t host_addr;
	struct hostent *host;

	host = gethostbyname(MQTT_BROKER);
	if (host == NULL)
	{
		DEBUG_PRINTF("MQTT CONNECT: gethostbyname failed\r\n");
		return -1;
	}

	/* Setup an empty client info structure */
	memset(&ci, 0, sizeof(ci));

	/* Minimal amount of information required is client identifier, so set it here */
	ci.client_id = MQTT_CLIENT_NAME;

	/* copy the network address to sockaddr_in structure */
	if ((host->h_addrtype == AF_INET) && (host->h_length == sizeof(ip_addr_t)))
	{
		memcpy(&host_addr, host->h_addr_list[0], sizeof(ip_addr_t));

		DEBUG_PRINTF("MQTT CONNECT: broker %s on port %d\r\n",
				inet_ntoa(host_addr),
				MQTT_BROKER_PORT);

		/* Initiate client and connect to server, if this fails immediately an error code is returned
		 otherwise client_mqtt_connection_cb will be called with connection result after attempting
		 to establish a connection with the server.
		 For now MQTT version 3.1.1 is always used */
		err = mqtt_client_connect(client, &host_addr, MQTT_BROKER_PORT, client_mqtt_connection_cb, 0, &ci);
	}
	else
	{
		DEBUG_PRINTF("MQTT CONNECT: gethostbyname returned invalid data\r\n");
		return -1;
	}

	/* For now just print the result code if something goes wrong */
	if(err != ERR_OK)
	{
		DEBUG_PRINTF("MQTT CONNECT return: %d\r\n", err);
	}

	return err;
}

static void client_mqtt_sub_request_cb(void *arg, err_t result)
{
	if (result != ERR_OK)
	{
		DEBUG_PRINTF("MQTT SUBSCRIBE result: %d\r\n", result);
	}
}

static void client_mqtt_incoming_publish_cb(void *arg, const char *topic, u32_t tot_len)
{
	struct sub_info *subInfo = (struct sub_info *)arg;

	DEBUG_PRINTF("MQTT RECIEVE: %s (%u)\r\n", topic, (unsigned int)tot_len);

	memset(subInfo, 0, sizeof(struct sub_info));

	/* Decode topic string into a user defined reference */
	if(strcmp(topic, TOPIC_CONTROL) == 0)
	{
		subInfo->topic = TOPIC_CONTROL;
	}
	tfp_printf("%d\r\n", __LINE__);
}

static void client_mqtt_incoming_data_cb(void *arg, const u8_t *data, u16_t len, u8_t flags)
{
	struct sub_info *subInfo = (struct sub_info *)arg;

	DEBUG_PRINTF("MQTT RECIEVE: %s (%d) flags %u\r\n", data, len, (unsigned int)flags);

	if (len > 0)
	{
		if (subInfo->topic)
		{
			if (len > sizeof(subInfo->message))
			{
				len = sizeof(subInfo->message);
			}
			subInfo->len = len;
			memcpy(subInfo->message, data, len);
		}
	}
	tfp_printf("%d\r\n", __LINE__);

	xTaskNotify(gx_TaskMonitor, (uint32_t)arg, eSetValueWithOverwrite);
	tfp_printf("%d\r\n", __LINE__);
}

/* Called when publish is complete either with sucess or failure */
static void client_mqtt_pub_request_cb(void *arg, err_t result)
{
	if (result != ERR_OK)
	{
		DEBUG_PRINTF("MQTT PUBLISH result: %d\r\n", result);
	}
}

err_t client_mqtt_publish(mqtt_client_t *client, void *arg)
{
	err_t err;
	u8_t qos = 2; /* QOS explained in MQTT specification */
	u8_t retain = 0; /* Do not retain the payload. */
	struct pub_info *pubInfo = (struct pub_info *)arg;

	err = mqtt_publish(client, pubInfo->topic, pubInfo->message, pubInfo->len, qos, retain, client_mqtt_pub_request_cb, arg);
	if (err != ERR_OK)
	{
		DEBUG_PRINTF("MQTT PUBLISH err: %d\r\n", err);
	}

	return err;
}

void client_mqtt(void)
{
	err_t err;
	mqtt_client_t mqtt;
	uint32_t ulNotifiedValue;
	struct pub_info *pinfo;
	struct sub_info sinfo;

	/* Connect once and stay connected. */
	memset(&mqtt, 0, sizeof(mqtt_client_t));
	client_mqtt_do_connect(&mqtt);

	while (mqtt_client_is_connected(&mqtt) == 0)
	{
		vTaskDelay(pdMS_TO_TICKS(MQTT_CONNECT_POLL_MS));
	}

	err = mqtt_subscribe(&mqtt, TOPIC_CONTROL, 1, client_mqtt_sub_request_cb, 0);

	/* Setup callback for incoming publish requests */
	mqtt_set_inpub_callback(&mqtt, client_mqtt_incoming_publish_cb, client_mqtt_incoming_data_cb, &sinfo);

	if(err != ERR_OK)
	{
		DEBUG_PRINTF("MQTT SUBSCRIBE err: %d\r\n", err);
	}

	while (1)
	{
		/* Wait for a notification from the monitor task.
		 * This holds a pointer to a string with the publish
		 * message.
		 */
		if (xTaskNotifyWait(0, 0, &ulNotifiedValue, pdMS_TO_TICKS(10)))
		{
			if (!net_is_link_up())
			{
				/* Link down. Suspend actions. */
				break;
			}

			if (mqtt_client_is_connected(&mqtt) == 0)
			{
				/* Server gone. Suspend actions. */
				break;
			}

			pinfo = (struct pub_info *)ulNotifiedValue;
			err = client_mqtt_publish(&mqtt, pinfo);
		}
	}

	mqtt_disconnect(&mqtt);
}

/**
 * This is the client task.
 *
 * @param [in]      pvParameters        Unused.
 */
void vTaskClient(void *pvParameters)
{
	(void) pvParameters;

	DEBUG_PRINTF("Task %s started.\r\n", __FUNCTION__);

	/* Initialise Ethernet module and LwIP service. */
	net_init(*NET_DEFAULT_IP_ADDR, *NET_DEFAULT_GW_ADDR, *NET_DEFAULT_MASK, NET_USE_DHCP, "FT90x_MQTT_Example",
			app_ethif_status_cb);

	if (xTaskCreate(vTaskConnect,
			"Connect",
			TASK_CONNECT_STACK_SIZE,
			NULL,
			TASK_CONNECT_PRIORITY,
			&gx_TaskConnect) != pdTRUE) {
		DEBUG_PRINTF("Connect Thread failed\r\n");
	}

	if (xTaskCreate(vTaskMonitor,
			"Monitor",
			TASK_MONITOR_STACK_SIZE,
			NULL,
			TASK_MONITOR_PRIORITY,
			&gx_TaskMonitor) != pdTRUE) {
		DEBUG_PRINTF("Monitor Thread failed\r\n");
	}

	/* Wait till network interface is up (IP address is valid in the case of DHCP) */
	if (!net_is_up())
	{
		DEBUG_PRINTF("Waiting for interface");
		while (!net_is_up())
		{
			vTaskDelay(pdMS_TO_TICKS(10));
			DEBUG_PRINTF(".");
		}
		DEBUG_PRINTF("\r\n");
	}

	while (1)
	{
		if (!net_is_link_up())
		{
			DEBUG_PRINTF("Waiting for link");
			while (!net_is_link_up())
			{
				vTaskDelay(pdMS_TO_TICKS(100));
				DEBUG_PRINTF(".");
			}
			DEBUG_PRINTF("\r\n");
		}

		DEBUG_PRINTF("Waiting for configuration");

		/* Wait for valid IP address */
		while(1)
		{
			vTaskDelay(pdMS_TO_TICKS(100));

			if (net_get_ip().addr != IPADDR_ANY)
			{
				DEBUG_PRINTF("\r\n");
				break;
			}
			DEBUG_PRINTF(".");
		}

		DEBUG_PRINTF("Starting\r\n");

		client_mqtt();
	}

	DEBUG_PRINTF("Task %s finished.\r\n", __FUNCTION__);
}
